/*-
 * Copyright (C)2019 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2019 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: const.c,v 1.1 2019/06/24 01:04:56 babolo Exp $"

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <babolo/BLINflag.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include "xml2pg.h"

static char *xtnm[XML2PG_TYNUM + 1] =
{ "none"               /*  0 */
, "ELEMENT_NODE"       /*  1 */
, "ATTRIBUTE_NODE"     /*  2 */
, "TEXT_NODE"          /*  3 */
, "CDATA_SECTION_NODE" /*  4 */
, "ENTITY_REF_NODE"    /*  5 */
, "ENTITY_NODE"        /*  6 */
, "PI_NODE"            /*  7 */
, "COMMENT_NODE"       /*  8 */
, "DOCUMENT_NODE"      /*  9 */
, "DOCUMENT_TYPE_NODE" /* 10 */
, "DOCUMENT_FRAG_NODE" /* 11 */
, "NOTATION_NODE"      /* 12 */
, "HTML_DOCUMENT_NODE" /* 13 */
, "DTD_NODE"           /* 14 */
, "ELEMENT_DECL"       /* 15 */
, "ATTRIBUTE_DECL"     /* 16 */
, "ENTITY_DECL"        /* 17 */
, "NAMESPACE_DECL"     /* 18 */
, "XINCLUDE_START"     /* 19 */
, "XINCLUDE_END"       /* 20 */
, "DOCB_DOCUMENT_NODE" /* 21 */
};

static char *ftnm[XML_WITH_LZMA + 1] =
{ "none"        /*  0 */
, "THREAD"      /*  1 */
, "TREE"        /*  2 */
, "OUTPUT"      /*  3 */
, "PUSH"        /*  4 */
, "READER"      /*  5 */
, "PATTERN"     /*  6 */
, "WRITER"      /*  7 */
, "SAX1"        /*  8 */
, "FTP"         /*  9 */
, "HTTP"        /* 10 */
, "VALID"       /* 11 */
, "HTML"        /* 12 */
, "LEGACY"      /* 13 */
, "C14N"        /* 14 */
, "CATALOG"     /* 15 */
, "XPATH"       /* 16 */
, "XPTR"        /* 17 */
, "XINCLUDE"    /* 18 */
, "ICONV"       /* 19 */
, "ISO8859X"    /* 20 */
, "UNICODE"     /* 21 */
, "REGEXP"      /* 22 */
, "AUTOMATA"    /* 23 */
, "EXPR"        /* 24 */
, "SCHEMAS"     /* 25 */
, "SCHEMATRON"  /* 26 */
, "MODULES"     /* 27 */
, "DEBUG"       /* 28 */
, "DEBUG_MEM"   /* 29 */
, "DEBUG_RUN"   /* 30 */
, "ZLIB"        /* 31 */
, "ICU"         /* 32 */
, "LZMA"        /* 33 */
};

static char *atnm[XML2PG_ATNUM + 1] =
{ "none"        /*  0 */
, "CDATA"       /*  1 */
, "ID"          /*  2 */
, "IDREF"       /*  3 */
, "IDREFS"      /*  4 */
, "ENTITY"      /*  5 */
, "ENTITIES"    /*  6 */
, "NMTOKEN"     /*  7 */
, "NMTOKENS"    /*  8 */
, "ENUMERATION" /*  9 */
, "NOTATION"    /* 10 */
};

static char *etnm[XML_ELEMENT_TYPE_ELEMENT + 1] =
{ "UNDEFINED"   /*  0 */
, "EMPTY"       /*  1 */
, "ANY"         /*  2 */
, "MIXED"       /*  3 */
, "ELEMENT"     /*  4 */
};

static char *ttnm[XML_ELEMENT_CONTENT_OR + 1] =
{ "none"        /*  0 */
, "PCDATA"      /*  1 */
, "ELEMENT"     /*  2 */
, "SEQ"         /*  3 */
, "OR"          /*  4 */
};

static char *otnm[XML_ELEMENT_CONTENT_PLUS + 1] =
{ "none"        /*  0 */
, "ONCE"        /*  1 */
, "OPT"         /*  2 */
, "MULT"        /*  3 */
, "PLUS"        /*  4 */
};

char *
xml2pg_namec(xml2pg_cf *options, int namet, int nameo) {
#   define blin_internal_flags (options->flags)
    char *rt = NULL;
    char *ro = NULL;
    char *r = NULL;

    if  ((namet < 0) || (namet > XML_ELEMENT_CONTENT_OR)) {
        if  (0 > asprintf(&rt, "%d", namet)) {
            ifBLIN_QW0("asprintf t");
            rt = NULL;
            goto out;
        }
    } else if (!(rt = strdup(ttnm[namet]))) {
        ifBLIN_QW0("strdup t");
        goto out;
    }
    if  ((nameo < 0) || (nameo > XML_ELEMENT_CONTENT_PLUS)) {
        if  (0 > asprintf(&ro, "%d", nameo)) {
            ifBLIN_QW0("asprintf o");
            ro = NULL;
            goto out;
        }
    } else if (!(ro = strdup(otnm[nameo]))) {
        ifBLIN_QW0("strdup o");
        goto out;
    }
    if  (0 > asprintf(&r, "%s:%s", rt, ro)) {
        ifBLIN_QW0("asprintf");
        r = NULL;
        goto out;
    }
    free(ro);
out:
    free(rt);
    return(r);
#   undef blin_internal_flags
}

char *
xml2pg_namex(xml2pg_cf *options, int name) {
#   define blin_internal_flags (options->flags)
    char *r = NULL;

    if  ((name <= 0) || (name > XML2PG_TYNUM)) {
        if  (0 > asprintf(&r, "%d", name)) {
            ifBLIN_QW0("asprintf");
            r = NULL;
            goto out;
        }
    } else if (!(r = strdup(xtnm[name]))) {
        ifBLIN_QW0("strdup");
        goto out;
    }
out:
    return(r);
#   undef blin_internal_flags
}

char *
xml2pg_namef(xml2pg_cf *options, int name) {
#   define blin_internal_flags (options->flags)
    char *r = NULL;

    if  ((name <= 0) || (name > XML_WITH_LZMA)) {
        if  (0 > asprintf(&r, "%d", name)) {
            ifBLIN_QW0("asprintf");
            r = NULL;
            goto out;
        }
    } else if (!(r = strdup(ftnm[name]))) {
        ifBLIN_QW0("strdup");
        goto out;
    }
out:
    return(r);
#   undef blin_internal_flags
}

char *
xml2pg_namea(xml2pg_cf *options, int name) {
#   define blin_internal_flags (options->flags)
    char *r = NULL;

    if  ((name <= 0) || (name > XML2PG_ATNUM)) {
        if  (0 > asprintf(&r, "%d", name)) {
            ifBLIN_QW0("asprintf");
            r = NULL;
            goto out;
        }
    } else if (!(r = strdup(atnm[name]))) {
        ifBLIN_QW0("strdup");
        goto out;
    }
out:
    return(r);
#   undef blin_internal_flags
}

char *
xml2pg_namee(xml2pg_cf *options, int name) {
#   define blin_internal_flags (options->flags)
    char *r = NULL;

    if  ((name < 0) || (name > XML_ELEMENT_TYPE_ELEMENT)) {
        if  (0 > asprintf(&r, "%d", name)) {
            ifBLIN_QW0("asprintf");
            r = NULL;
            goto out;
        }
    } else if (!(r = strdup(etnm[name]))) {
        ifBLIN_QW0("strdup");
        goto out;
    }
out:
    return(r);
#   undef blin_internal_flags
}
