/*-
 * Copyright (C)2019 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2019 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: xml2pg.c,v 1.17 2019/06/26 08:05:10 babolo Exp $"

#define MIFE_COMPAT 5

#include <sys/types.h>
#include <sys/event.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <sys/time.h>
#include <sysexits.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <getCGIparm.h>
#include <mife.h>
#include "xml2pg.h"

static u_char *
forcopy(xml2pg_cf *options, const u_char *n) {
#   define blin_internal_flags (options->flags)
    int     ex = EX_OK;
    u_char *m;
    char   *u;

    if  (!n) {
        if  (!(m = (u_char*)strdup("\\N"))) ifBLIN_QW0("strdup \\N");
    } else if (!(m = (u_char*)(u = strdup((char*)n)))) {
        ifBLIN_QW0("strdup %s", n);
    } else if ((ex = getCGIparmpass3(0, (u_char*)"copy", &m))) {
        ifBLIN_QW0("getCGIparmpass3 %s", m);
        if  ((u_char*)u != m) free(u);
        free(m);
        m = NULL;
    } else {
        if  ((u_char*)u != m) free(u);
    }
    return(m);
#   undef blin_internal_flags
}

static int
xobjdump(xml2pg_cf *options, xmlNode *xobj) {
#   define blin_internal_flags (options->flags)
    int          ex = EX_OK;
    xmlBufferPtr buffer;
    int          l = 0;
    u_char      *n0;

    if  (options->flags & XML2PG_DUMP) {
        if  (options->flags & XML2PG_DUUP) l = 1;
        if  (!(buffer = xmlBufferCreate())) {
            ifBLIN_QW0("xmlBufferCreate");
            ex = EX_OSERR;
            goto out;
        }
        if  (0 > xmlNodeDump(buffer, options->xdoc, xobj, l, l)) {
            ifBLIN_QW0("xmlNodeDump");
            ex = EX_OSERR;
            goto out;
        }
        if  (!(n0 = forcopy(options, xmlBufferContent(buffer)))) {
            ifBLIN_QW0("forcopy n0");
            ex = EX_OSERR;
            goto out;
        }
        printf("\t%s", n0);
        free(n0);
    }
out:
    return(ex);
#   undef blin_internal_flags
}

static int
commo( xml2pg_cf    *options
     , xmlNode      *xobj
     , int           ex   /*        */
     , int           exx  /*                   */
     , int           shift
     ) {
#   define blin_internal_flags (options->flags)
    int      nex = EX_OK;

    ifBLIN_QX2("+%08X %u->%u", options->flags, ex, exx);
    if  (~options->flags & XML2PG_SHFT) shift = 0;
    if  (0 > ex) {
        printf("%d\t\\N", exx);                                        /* 0 (num) and 1 (up) columns */
    } else {
        printf("%d\t%d", exx, ex);                                     /* 0 (num) and 1 (up) columns */
    }
    if  (options->flags & XML2PG_NAME) {
        printf("\t%s", xml2pg_namex(options, xobj->type));                        /* 2 (type) column */
    } else {
        printf("\t%d", shift + ((xobj->type < XML2PG_TYNUM) ? xobj->type : 0));   /* 2 (type) column */
    }
    ifBLIN_QX2("- %d", nex);
    return(nex);
#   undef blin_internal_flags
}

static int
namen( xml2pg_cf    *options
     , xmlNode      *xobj
     , int           ex   /*        */
     , int           exx  /*                   */
     , const u_char *ns
     ) {
#   define blin_internal_flags (options->flags)
    int      nex = EX_OK;
    char    *nodename;

    ifBLIN_QX2("+%08X %u->%u", options->flags, ex, exx);
    if  (  (options->flags & XML2PG_FULL)
        && !!ns
        ) {
        if  (0 > asprintf(&nodename, "%s:%s", ns, xobj->name)) {
            ifBLIN_QW0("asprintf nodename");
            nex = EX_UNAVAILABLE;
            goto out;
        }
    } else if (!xobj->name) {
        if  (!(nodename = strdup("\\N"))) {
            ifBLIN_QW0("strdup nodename");
            nex = EX_OSERR;
            goto out;
        }
    } else if (!(nodename = strdup((const char *)xobj->name))) {
        ifBLIN_QW0("strdup nodename");
        nex = EX_OSERR;
        goto out;
    }
    printf("\t%s", nodename);                                                     /* 3 (name) column */
    free(nodename);
out:
    ifBLIN_QX2("- %d", nex);
    return(nex);
#   undef blin_internal_flags
}

static int
dentry( xml2pg_cf    *options
      , void         *x
      , int           ex   /*        */
      , int           exx  /*                   */
      , const u_char *ns
      , const u_char *name
      , const u_char *val
      , const u_char *vval
      ) {
#   define blin_internal_flags (options->flags)
    int      eex = EX_OK;
    u_char  *n0 = NULL;

    ifBLIN_QX2("+%08X %u->%u", options->flags, ex, exx);
    if  (!name) {
        if  (!!(eex = namen(options, x, ex, exx, ns))) {
            ifBLIN_QW0("namen");
            goto out;
        }
    } else {
        if  (!(n0 = forcopy(options, name))) {
            ifBLIN_QW0("forcopy name");
            eex = EX_OSERR;
            goto out;
        }
        printf("\t%s", n0);
        free(n0);
        n0 = NULL;
    }
    if  (!(n0 = forcopy(options, val))) {
        ifBLIN_QW0("forcopy val");
        eex = EX_OSERR;
        goto out;
    }
    printf("\t%s", n0);                                                          /* 4 (value) column */
    free(n0);
    n0 = NULL;
    if  (options->flags & XML2PG_COPZ) {
        if  (!(n0 = forcopy(options, vval))) {
            ifBLIN_QW0("forcopy vval");
            eex = EX_OSERR;
            goto out;
        }
        printf("\t%s", n0);
        free(n0);
        n0 = NULL;
    }
    if  (!!xobjdump(options, (xmlNode *)x)) {
        ifBLIN_QW0("xobjdump xns");
        eex = EX_OSERR;
        goto out;
    }
    printf("\n");
out:
    ifBLIN_QX2("- %d", eex);
    return(eex);
#   undef blin_internal_flags
}

int
content( xml2pg_cf         *options
       , xmlElementContent *xcont
       , int                ex   /*        */
       , int                exx  /*                   */
       , int                shift
       ) {
#   define blin_internal_flags (options->flags)
    char    *n0 = NULL;
    int      nex;

    ifBLIN_QX2("+%08X %u->%u %i %"BLIN_X, options->flags, ex, exx, XML2PG_CONTN, xcont);
    nex = exx;
    if  (~options->flags & XML2PG_SHFT) shift = 0;
    if  (0 > ex) {
        printf("%d\t\\N", exx);                                        /* 0 (num) and 1 (up) columns */
    } else {
        printf("%d\t%d", exx, ex);                                     /* 0 (num) and 1 (up) columns */
    }
    if  (options->flags & XML2PG_NAME) {
        printf("\tCONTENT");                                                      /* 2 (type) column */
    } else {
        printf("\t%d", shift + XML2PG_CONTN);                                     /* 2 (type) column */
    }
    if  (  (options->flags & XML2PG_FULL)
        && !!xcont->prefix
        ) {
        if  (0 > asprintf(&n0, "%s:%s", xcont->prefix, xcont->name)) {
            ifBLIN_QW0("asprintf n0");
            ex = -EX_UNAVAILABLE;
            goto out;
        }
    } else if (!xcont->name) {
        if  (!(n0 = strdup("\\N"))) {
            ifBLIN_QW0("strdup n0");
            ex = -EX_OSERR;
            goto out;
        }
    } else if (!(n0 = strdup((const char *)xcont->name))) {
        ifBLIN_QW0("strdup n0");
        ex = -EX_OSERR;
        goto out;
    }
    printf("\t%s", n0);                                                           /* 3 (name) column */
    free(n0);
    n0 = xml2pg_namec(options, xcont->type, xcont->ocur);
    printf("\t%s", n0);                                                          /* 4 (value) column */
    free(n0);
    if  (options->flags & XML2PG_COPZ) printf("\t\\N");
    if  (options->flags & XML2PG_DUMP) printf("\t\\N");
    printf("\n");
    if  ((xcont->c1) && (0 > (exx = content(options, xcont->c1, nex, ++exx, 0)))) {
        ifBLIN_QW0("content c1");
        ex = -exx;
        goto out;
    }
    if  ((xcont->c2) && (0 > (exx = content(options, xcont->c2, nex, ++exx, 0)))) {
        ifBLIN_QW0("content c2");
        ex = -exx;
        goto out;
    }
    ex = exx;
out:
    ifBLIN_QX2("-%08X %u->%u", options->flags, ex, ++exx);
    return(ex);
#   undef blin_internal_flags
}

int
xml2pg_dump( xml2pg_cf    *options
           , void         *x
           , int           ex   /*        */
           , int           exx  /*                   */
           , int           shift
           ) {
#   define blin_internal_flags (options->flags)
    xmlNode *xobj = x;
    int      eex = EX_OK;
    char    *n0 = NULL;
    int      nex;

    ifBLIN_QX2("+%08X %u->%u %i %"BLIN_X, options->flags, ex, exx, xobj->type, xobj);
    ifBLIN_QX3("%"BLIN_X" %"BLIN_X" %"BLIN_X" %"BLIN_X" %"BLIN_X" %"BLIN_X
              , xobj->children, xobj->last, xobj->parent, xobj->next, xobj->prev, xobj->doc
              );
    nex = exx;
    if  (!!(eex = commo(options, x, ex, exx, shift))) {
        ifBLIN_QW0("commo");
        ex = -eex;
        goto out;
    }
    switch(xobj->type) {
    case XML_ELEMENT_NODE:       //  1
    case XML_TEXT_NODE:          //  3
    case XML_CDATA_SECTION_NODE: //  4
    case XML_ENTITY_REF_NODE:    //  5
    case XML_ENTITY_NODE:        //  6
    case XML_PI_NODE:            //  7
    case XML_COMMENT_NODE:       //  8
    case XML_DOCUMENT_TYPE_NODE: // 10
    case XML_DOCUMENT_FRAG_NODE: // 11
    case XML_NOTATION_NODE:      // 12
    case XML_HTML_DOCUMENT_NODE: // 13
    case XML_XINCLUDE_START:     // 19
    case XML_XINCLUDE_END:       // 20
    case XML_DOCB_DOCUMENT_NODE: // 21
        {   xmlChar *cur = NULL;

            cur = xmlNodeListGetString(options->xdoc, xobj, 1);
            if  (!!(eex = dentry( options
                                , x
                                , ex
                                , exx
                                , !!xobj->ns ? xobj->ns->prefix : NULL
                                , NULL
                                , xobj->content
                                , cur
                )  )            ) {
                ifBLIN_QW0("dentry");
                ex = -eex;
                goto out;
            }
            xmlFree(cur);
            for (xmlAttr *attr = xobj->properties; attr; attr = attr->next) {
                ifBLIN_QX3("%i %i %i %"BLIN_X, ex, exx, attr->type, xobj);
                if  (0 > (exx = xml2pg_dump(options, attr, nex, ++exx, 0))) {
                    ifBLIN_QW0("xml2pg_dump attr");
                    ex = -exx;
                    goto out;
            }   }
#if 0
            for (xmlNs *xns = xobj->ns; xns; xns = xns->next) {
                if  (0 > (exx = xml2pg_dump(options, xns, nex, ++exx, 64))) {
                    ifBLIN_QW0("xml2pg_dump in xobj->ns");
                    ex = -ex;
                    goto out;
            }   }
#endif
            for (xmlNs *xns = xobj->nsDef; xns; xns = xns->next) {
                ifBLIN_QX3("%i %i %i %"BLIN_X, ex, exx, xns->type, xobj);
                if  (0 > (exx = xml2pg_dump(options, xns, nex, ++exx, 96))) {
                    ifBLIN_QW0("xml2pg_dump in xobj->nsDef");
                    ex = -exx;
                    goto out;
            }   }
            for (xobj = xobj->children; xobj; xobj = xobj->next) {
                ifBLIN_QX3("%i %i %i %"BLIN_X, ex, exx, xobj->type, xobj);
                if  (0 > (exx = xml2pg_dump(options, xobj, nex, ++exx, 0))) {
                    ifBLIN_QW0("xml2pg_dump in xobj");
                    ex = -exx;
                    goto out;
        }   }   }
        break;
    case XML_ATTRIBUTE_NODE:     //  2
        {   xmlAttr *attr = x;
            xmlChar *val  = xmlNodeListGetRawString(attr->doc, attr->children, 1);

            n0 = xml2pg_namea(options, attr->atype);
            if  (!!(eex = dentry(options, x, ex, exx, NULL, NULL, val, (u_char *)n0))) {
                ifBLIN_QW0("dentry");
                ex = -eex;
                goto out;
            }
            for (xmlNs *xns = attr->ns; xns; xns = xns->next) {
                ifBLIN_QX3("%i %i %i %"BLIN_X, ex, exx, xobj->type, xobj);
                if  (0 > (exx = xml2pg_dump(options, xns, nex, ++exx, 0))) {
                    ifBLIN_QW0("xml2pg_dump in xns");
                    ex = -exx;
                    goto out;
            }   }
            for (xobj = xobj->children; xobj; xobj = xobj->next) {
                ifBLIN_QX3("%i %i %i %"BLIN_X, ex, exx, xobj->type, xobj);
                if  (0 > (exx = xml2pg_dump(options, xobj, nex, ++exx, 0))) {
                    ifBLIN_QW0("xml2pg_dump in xobj");
                    ex = -exx;
                    goto out;
        }   }   }
        break;
    case XML_DOCUMENT_NODE:      //  9
        {   xmlDoc    *xdoc = x;

            if  (!!(eex = dentry(options, x, ex, exx, NULL, NULL, xdoc->URL, xdoc->version))) {
                ifBLIN_QW0("dentry");
                ex = -eex;
                goto out;
            }
#if 0
            for (xmlDtd *xdtd = xdoc->intSubset; xdtd; xdtd = (xmlDtd*)xdtd->next) {
                ifBLIN_QX3("%i %i %i %"BLIN_X, ex, exx, xdtd->type, xdtd);
                if  (0 > (exx = xml2pg_dump(options, xdtd, nex, ++exx, 0))) {
                    ifBLIN_QW0("xml2pg_dump in xdoc");
                    ex = -exx;
                    goto out;
            }   }
            if  (xdoc->intSubset != xdoc->extSubset) {
                for (xmlDtd *xdtd = xdoc->extSubset; xdtd; xdtd = (xmlDtd*)xdtd->next) {
                    ifBLIN_QX3("%i %i %i %"BLIN_X, ex, exx, xdtd->type, xdtd);
                    if  (0 > (exx = xml2pg_dump(options, xdtd, nex, ++exx, 0))) {
                        ifBLIN_QW0("xml2pg_dump in xdoc");
                        ex = -exx;
                        goto out;
            }   }   }
#endif
            for (xmlNode *xobj = xdoc->children; xobj; xobj = xobj->next) {
                ifBLIN_QX3("%i %i %i %"BLIN_X, ex, exx, xobj->type, xobj);
                if  (0 > (exx = xml2pg_dump(options, xobj, nex, ++exx, 0))) {
                    ifBLIN_QW0("xml2pg_dump in xdoc");
                    ex = -exx;
                    goto out;
        }   }   }
        break;
    case XML_DTD_NODE:           // 14
        {   xmlDtd    *xdtd = x;

            if  (!!(eex = dentry(options, x, ex, exx, NULL, NULL, xdtd->SystemID, xdtd->ExternalID))) {
                ifBLIN_QW0("dentry");
                ex = -eex;
                goto out;
            }
            for (xobj = xdtd->children; xobj; xobj = xobj->next) {
                ifBLIN_QX3("%i %i %i %"BLIN_X, ex, exx, xobj->type, xobj);
                if  (0 > (exx = xml2pg_dump(options, xobj, nex, ++exx, 0))) {
                    ifBLIN_QW0("xml2pg_dump in xdoc");
                    ex = -exx;
                    goto out;
        }   }   }
        break;
    case XML_ELEMENT_DECL:       // 15
        {   xmlElement *xele = x;

            n0 = xml2pg_namee(options, xele->etype);
            if  (!!(eex = dentry(options, x, ex, exx, xele->prefix, NULL, (u_char *)n0, NULL))) {
                ifBLIN_QW0("dentry");
                ex = -eex;
                goto out;
            }
            if  (!!xele->content && (0 > (exx = content(options, xele->content, nex, ++exx, 0)))) {
                ifBLIN_QW0("content");
                ex = -eex;
                goto out;
            }
            for (xmlAttribute *attr = xele->attributes; attr; attr = (xmlAttribute *)attr->next) {
                ifBLIN_QX3("%i %i %i %"BLIN_X, ex, exx, attr->type, attr);
                if  (0 > (exx = xml2pg_dump(options, attr, nex, ++exx, 0))) {
                    ifBLIN_QW0("xml2pg_dump attr");
                    ex = -exx;
                    goto out;
        }   }   }
        break;
    case XML_ATTRIBUTE_DECL:     // 16
        {   xmlAttribute *attr = x;

            n0 = xml2pg_namea(options, attr->atype);
            if  (!!(eex = dentry(options, x, ex, exx, NULL, NULL, (u_char *)n0, NULL))) {
                ifBLIN_QW0("dentry");
                ex = -eex;
                goto out;
        }   }
        break;
    case XML_ENTITY_DECL:        // 17
        {   xmlEntity *xent = x;

            if  (!!(eex = dentry(options, x, ex, exx, NULL, NULL, xent->SystemID, xent->ExternalID))) {
                ifBLIN_QW0("dentry");
                ex = -eex;
                goto out;
            }
#if 0
            printf("etype=%d orig=%s~ content=%s~ URI=%s~\n", xent->etype, xent->orig, xent->content, xent->URI);
#endif
        }
        break;
        /* XXXX xmlChar       * xent->orig */
        /* XXXX xmlChar       * xent->content */
        /* XXXX xmlEntityType   xent->etype */
        /* XXXX const xmlChar *	xent->URI */
    case XML_NAMESPACE_DECL:     // 18
        {   xmlNs    *xns = x;

            if  (!!(eex = dentry(options, x, ex, exx, NULL, xns->prefix, xns->href, NULL))) {
                ifBLIN_QW0("dentry");
                ex = -eex;
                goto out;
        }   }
        break;
    default: {
            char *t;

            t = (options->flags & XML2PG_NAME) ? "none" : "0";
            printf( (options->flags & XML2PG_COPZ)
                  ? "%d\t%d\t%s\t%s\t\\N\t\\N\n"
                  : "%d\t%d\t%s\t%s\t\\N\n"
                  , ++exx
                  , ex
                  , t
                  , xobj->name
                  );
        }
        break;
    }
    ex = exx;
out:
    free(n0);
    ifBLIN_QX2("-%08X %u->%u", options->flags, ex, ++exx);
    return(ex);
#   undef blin_internal_flags
}

static void
usage(xml2pg_cf *options, int ex) {
    const char *fl[XML2PG_FLIST];
    int         tifl;
    int         ifl;
    int         i;
    char       *s;

    fprintf( stderr
           , "xml2pg @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
             "    -b        - do not create blank nodes\n"
             "    -c[c]     - COPY object recursive/plain\n"
             "    -d[d]     - recursive dump node to last column\n"
             "    -f[f]     - name with namespace\n"
             "    -n        - symbolic type names\n"
             "    -q        - quiet\n"
             "    -s        - shift type numbers\n"
             "    -v        - verbose\n"
             " Feature list:\n"
           )
    ;
    if  (0 > (tifl = xmlGetFeaturesList(&ifl, fl))) {
        ifBLIN_QX0("xmlGetFeaturesList");
    } else {
        for (i = 0; i < ifl; ++i) {
            fprintf(stderr, "%s\n", fl[i]);
    }   }
    fprintf(stderr, " Has features:\n");
    for (int i = 1; i <= XML_WITH_LZMA; ++i) {
        s = xml2pg_namef(options, i);
        fprintf(stderr, "    XML_WITH_%s	%d\n", s, xmlHasFeature(i));
        free(s);
    }
    fprintf(stderr, " List of node types:\n");
    for (i = 1; i <= XML2PG_TYNUM; ++i) {
        s = xml2pg_namex(options, i);
        fprintf(stderr, "    %2d XML_%s\n", i, s);
        free(s);
    }
    fprintf(stderr, " List of attribute types:\n");
    for (i = 1; i <= XML2PG_ATNUM; ++i) {
        s = xml2pg_namea(options, i);
        fprintf(stderr, "    %2d XML_ATTRIBUTE_%s\n", i, s);
        free(s);
    }
    exit(ex);
}

int
main(int argc, char **argv) {
#   define blin_internal_flags (options->flags)
    const char          *flar = "bcdfnqsv=:?";
    int                  ex = EX_OK;
    babolo_opts         *bos = NULL;
    xml2pg_cf           *options;
    off_t                xoffset;
    ssize_t              xsize;
    xmlParserCtxtPtr     ctxt;
    int                  xnum;
    mife_descriptor     *mdi;
    const char          *o;
    int                  c;

    LIBXML_TEST_VERSION;
    if  (!(options = calloc(1, sizeof(*options)))) {
        warnx("Malloc failed #1");
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    options->flags = BLIN_BIT0;
    if  (!(bos = babolo_openopts(Bpars_NONU | Bpars_SYME, 'Z'))) {
        ifBLIN_QW0("babolo_openopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (babolo_setopts(bos, Bpars_NONU, argc, argv, flar)) {
        ifBLIN_QW0("babolo_setopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    while (0 < (c = babolo_getopts(bos))) {
        switch (c) {
        case 'b': options->flags |= XML2PG_BLNK;
                ; break;
        case 'c': options->flags |= (options->flags & XML2PG_COPY) << 1;
                ; options->flags |= XML2PG_COPY;
                ; break;
        case 'd': options->flags |= (options->flags & XML2PG_DUMP) << 1;
                ; options->flags |= XML2PG_DUMP;
                ; break;
        case 'f': options->flags |= (options->flags & XML2PG_FULL) << 1;
                ; options->flags |= XML2PG_FULL;
                ; break;
        case 'n': options->flags |= XML2PG_NAME;
                ; break;
        case 'q': options->flags &= ~BLIN_MASK;
                ; break;
        case 's': options->flags |= XML2PG_SHFT;
                ; break;
        case 'v': BLIN_VERBOSE(options->flags);
                ; break;
        case '=': if  (!(o = babolo_getoptsarg(bos))) {
                ;     ifBLIN_QW0("No -%c arg", bos->c);
                ;     ex = EX_USAGE;
                ;     goto out;
                ; }
                ; switch(o[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);           exit(EX_OK);
                  case '1': babolo_optest(flar);               exit(EX_OK);
                  case '2': babolo_optext(bos);                exit(EX_OK);
                  case '3': babolo_dumpopts(bos);              exit(EX_OK);
                  }
                ; /* FALLTHROUGH */
        case '?': usage(options, EX_OK);
        default : ifBLIN_QX0("Illegal flag -%c", c);
                ; usage(options, EX_USAGE);
    }   }
    options->subflags = MIFE_SLID
                      | ( BLIN_MASK
                        & ((BLIN_GEN1 & options->flags) | (options->flags >> 3)
                        ) )
    ;
    if  (!(mdi = mife_init(options->subflags))) {
        ifBLIN_QW0("mife_init stdin");
        ex = EX_IOERR;
        goto out;
    }
    if  (0 > mife_ctlfdsc(mdi, fileno(stdin))) {
        ifBLIN_QW0("mife_ctlfdsc stdin");
        ex = EX_IOERR;
        goto out;
    }
    xoffset = 0;
    xnum = 0;
    if  (options->flags & XML2PG_BLNK) xmlKeepBlanksDefault(0);
    for (int i = 0; ; ++i, xoffset += xsize) {
        if  (0 > (xsize = mife_read(mdi, 0, xoffset))) {
            ifBLIN_QW0("mife_read stdin");
            ex = EX_IOERR;
            goto out;
        }
        if  (!xsize) {
            ifBLIN_QX0("null stdin");
            ex = EX_DATAERR;
            goto out;
        }
        if  (!(o = mife_get(mdi, xoffset))) {
            ifBLIN_QW0("mife_get stdin");
            ex = EX_IOERR;
            goto out;
        }
        if  (!i) {
            if  (!(ctxt = xmlCreatePushParserCtxt(NULL, NULL, o, xsize, NULL))) {
                ifBLIN_QW0("xmlCreatePushParserCtxt");
                errno = EDOOFUS;
                ex = EX_SOFTWARE;
                goto out;
            }
        } else {
            if  (!!(ex = xmlParseChunk(ctxt, o, xsize, 0))) {
                ifBLIN_QX0("xmlParseChunk %d", ex);                     /* XXXX Enum xmlParserErrors */
                errno = EDOOFUS;
                ex = EX_SOFTWARE;
                goto out;
        }   }
        if  (mdi->flags & MIFE_EOFL) {
            if  (!!(ex = xmlParseChunk(ctxt, o, 0, 0))) {
                ifBLIN_QX0("xmlParseChunk %d", ex);                     /* XXXX Enum xmlParserErrors */
                errno = EDOOFUS;
                ex = EX_SOFTWARE;
                goto out;
            }
            break;
    }   }
    if  (!ctxt->wellFormed) {
        ifBLIN_QW0("Failed to parse");
        errno = EFTYPE;
        ex = EX_SOFTWARE;
        goto out;
    }
    options->xdoc = ctxt->myDoc;
    xmlFreeParserCtxt(ctxt);
    if  (options->flags & XML2PG_COPY) {
        if  (0 > (xnum = xml2pg_dump(options, options->xdoc, -1, xnum, 0))) {
            ifBLIN_QX0("xml2pg_dump");
            ex = -xnum;
            goto out;
        } else {
            ex = 0;
    }   }
    xmlFreeDoc(options->xdoc);
    xmlCleanupParser();
    xmlMemoryDump();
out:
    exit(ex);
#   undef blin_internal_flags
}
